#include "onwire.h"
#include "settings.h"
#include "sbit.h"
#include <util/crc16.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <stdbool.h>

onwi_t onwi;

static void onwi_start(uint8_t delay)
{
  onwi.delay = delay;
  onwi.state = OW_START;
  onwi.pbuf = onwi.buf;
  TCCR0A = 0;                   // Mode 0
  TCNT0 = 256 - OW_BIT_SLOT;
  TCCR0B = 1<<CS01 | 1<<CS00;   // F_CPU / 64
  TIMSK0 = 1<<TOIE0;
}

void onwi_measure(void)
{
  onwi.buf[0] = ONWI_SKIP_ROM;
  onwi.buf[1] = ONWI_CONVERT_T;
  onwi.wr_cnt = 2;
  onwi.rd_cnt = 0;
  onwi_start(0);
}

void onwi_read_temp(uint8_t delay)
{
  onwi.buf[0] = ONWI_SKIP_ROM;
  onwi.buf[1] = ONWI_READ_T;
  onwi.wr_cnt = 2;
  onwi.rd_cnt = 9;
  onwi_start(delay);
}

void onwi_read_rom(void)
{
  onwi.buf[0] = ONWI_READ_ROM;
  onwi.wr_cnt = 1;
  onwi.rd_cnt = 8;
  onwi.state = OW_START;
  onwi_start(0);
}

uint8_t onwi_crc(uint8_t cnt)
{
  uint8_t crc = 0;
  uint8_t* pbuf = onwi.buf;
  while (cnt--)
    crc = _crc_ibutton_update(crc, *pbuf++);
  return crc;
}

ISR(TIMER0_OVF_vect)
{
  if (ONWI_oe && !ONWI_out)             // if low
  {
    ONWI_oe = 0;                        // pin = tristate
    _delay_us(5);                       // min high time
  }
  uint8_t* pbuf = onwi.pbuf;            // for faster access
  switch (++onwi.state)
  {
    case OW_RES_LO:                     // 480us low
      if (onwi.delay--)                 // n * 100ms
        onwi.state = OW_DELAY;
      else
      {
        ONWI_out = 0;
        ONWI_oe = 1;                    // pin = low
      }
    case OW_RES_HI:                     // 480us high
    default:                            // n * 480µs
      TCNT0 = 256 - OW_RESET_SLOT;
      return;
    case OW_WR_DONE:                    // 1. command byte finished
      pbuf++;
      onwi.state = OW_WR_BIT0;
    case OW_WR_BIT0:                    // write bit 0
      if (onwi.wr_cnt--)
      {
    case OW_WR_BIT1:                    // write bit 1
    case OW_WR_BIT2:
    case OW_WR_BIT3:
    case OW_WR_BIT4:
    case OW_WR_BIT5:
    case OW_WR_BIT6:
    case OW_WR_BIT7:                    // write bit 7
        ONWI_oe = 1;                    // pin = low
        _delay_us(1);
        if (*pbuf & 1)
          ONWI_oe = 0;                  // pin = tristate
        *pbuf >>= 1;
        break;
      }
      ONWI_out = 1;                     // parasite power on
      ONWI_oe = 1;
      pbuf = onwi.buf - 1;
    case OW_RD_DONE:
      pbuf++;
      onwi.state = OW_RD_BIT0;
      if (onwi.rd_cnt--)
      {
        ONWI_oe = 0;                    // parasite power off
        ONWI_out = 0;
    case OW_RD_BIT1:                    // read bit 1
    case OW_RD_BIT2:
    case OW_RD_BIT3:
    case OW_RD_BIT4:
    case OW_RD_BIT5:
    case OW_RD_BIT6:
    case OW_RD_BIT7:                    // read bit 7
        ONWI_oe = 1;                    // pin = low
        _delay_us(1);
        ONWI_oe = 0;                    // pin = tristate
        _delay_us(5);
        *pbuf >>= 1;
        if (ONWI_in)
          *pbuf |= 0x80;
        break;
      }
      TCCR0B = 0;                       // stop T0
      return;
  }
  onwi.pbuf = pbuf;
  TCNT0 = 256 - OW_BIT_SLOT;
}
